/**
 * indexer
 * Copyright(c) 2011 Jesus A. Domingo <jesus.domingo@gmail.com>
 * MIT Licensed
 */

/**
 * Constructor
 */
function Indexer(options) {
  
  this._values = [];
  this._keymap = [];
  
  if (typeof options == 'object') {

    this.name = options.name;
    this._type = (options.type === 'numeric' ? 'numeric' : 'string');
    this._length = parseInt(options.length) || 0

  } else {
    
    this.name = options;
    this._type = 'string';
    this._length = 0;
  
  }
}

/**
 * Search the index for all keys having this value
 */
Indexer.prototype.search = function (value) {
  
  if (this._type == 'string' && this._length != 0) {
    value = value.substr(0, this._length);
  }
  
  var meta = this._getValuePosition(value);
  
  if (meta.found) {
    return this._keymap[meta.position];
  }
  
  return [];
};

/**
 * Enumerate all keys in order
 */
Indexer.prototype.all = function (limit, offset) {
  
  if (typeof limit != 'number') {
    limit = 0;
  }

  if (typeof offset != 'number') {
    offset = 0;
  }

  offset = (offset < 0 ? 0 : offset);
  limit = (limit < 0 ? 0 : limit);

  var sieve = {},
      klist = null;
  
  for (var i = 0, n = this._keymap.length; i < n; ++i) {
    klist = this._keymap[i];
    for (var j = 0, m = klist.length; j < m; ++j) {
      sieve[klist[j]] = 1;
    }
  }

  var keys = Object.keys(sieve);

  return (limit > 0 ? keys.slice(offset, offset + limit) : keys);
};

/**
 * Add a value and its key to the index
 */
Indexer.prototype.add = function (value, key) {

  if (this._type == 'string' && this._length != 0) {
    value = value.substr(0, this._length);
  }

  var meta = this._getValuePosition(value);

  if (meta.found) {
    this._keymap[meta.position].push(key);
  } else {
    this._values.splice(meta.position, 0, value);
    this._keymap.splice(meta.position, 0, [key]);
  }
};

/**
 * This removes a key from the index
 */
Indexer.prototype.remove = function (key) {
  for (var i = 0, n = this._keymap.length, p = 0; i < n; ++i) {
    p = this._keymap[i].indexOf(key);
    if (p >= 0) {
      delete this._keymap[i].splice(p, 1);
    }
  }
};

/**
 * Flushes keys for an indexed value
 */
Indexer.prototype.flush = function (value) {

  if (this._type == 'string' && this._length != 0) {
    value = value.substr(0, this._length);
  }

  var p = this._values.indexOf(value);
  
  if (p >= 0) {
    delete this._values.splice(p, 1);
    delete this._keymap.splice(p, 1);
  }
};

/**
 * Do a binary search on the _values list for the position of existing
 * values or where a non-existing value should be
 */
Indexer.prototype._getValuePosition = function (value) {

  var values = this._values;

  var low = 0,
      high = values.length - 1,
      middle = 0,
      comparison = 0,
      offset = 0,
      found = false;

  if (this._type == 'string') {
    value = value.toLowerCase();
  }

  while (low <= high) {

    middle = Math.floor((low + high) / 2);

    if (this._type == 'string') {
      comparison = values[middle].toLowerCase().localeCompare(value);
    } else {
      comparison = values[middle] - value;
    }

    if (comparison != 0) {
      if (comparison < 0) {
        low = middle + 1;
        offset = 1;
      } else {
        high = middle - 1;
        offset = 0;
      }
      continue;
    }

    found = true;
    offset = 0;
    
    break;
  }

  return { position: middle + offset, found: found };
};

/**
 * Create an index table
 */
exports.create = function (options) {
  return new Indexer(options);
};
